from typing import Sequence

from .stubs import DF17_BDS05, DF17_BDS65, DF20, DF21

def aircraft_information(
    icao24: str, registration: None | str = None
) -> dict[str, str]: ...
def decode_1090(msg: str) -> list[int]: ...
def decode_1090_with_reference(
    msg: str, reference: tuple[float, float]
) -> list[int]: ...
def decode_1090_vec(msgs: Sequence[Sequence[str]]) -> list[int]: ...
def decode_1090t_vec(
    msgs: Sequence[Sequence[str]],
    ts: Sequence[Sequence[float]],
    reference: None | tuple[float, float] = None,
) -> list[int]: ...
def decode_flarm(
    msg: str, timestamp: int, reflat: float, reflon: float
) -> list[int]: ...
def decode_flarm_vec(
    msgs: Sequence[Sequence[str]],
    ts: Sequence[Sequence[int]],
    reflat: Sequence[Sequence[float]],
    reflon: Sequence[Sequence[float]],
) -> list[int]: ...
def decode_bds05(mgs: str) -> DF17_BDS05: ...
def decode_bds10(mgs: str) -> DF20 | DF21: ...
def decode_bds17(mgs: str) -> DF20 | DF21: ...
def decode_bds18(mgs: str) -> DF20 | DF21: ...
def decode_bds19(mgs: str) -> DF20 | DF21: ...
def decode_bds20(mgs: str) -> DF20 | DF21: ...
def decode_bds21(mgs: str) -> DF20 | DF21: ...
def decode_bds30(mgs: str) -> DF20 | DF21: ...
def decode_bds40(mgs: str) -> DF20 | DF21: ...
def decode_bds44(mgs: str) -> DF20 | DF21: ...
def decode_bds45(mgs: str) -> DF20 | DF21: ...
def decode_bds50(mgs: str) -> DF20 | DF21: ...
def decode_bds60(mgs: str) -> DF20 | DF21: ...
def decode_bds65(mgs: str) -> DF17_BDS65: ...
